from .bounded_2d_kde import Bounded_2d_kde as bounded_2d_kde
from . import utilities as utils
import numpy as np
import corner
import matplotlib.pyplot as plt

def get_contour(costs, spins, level=0.9, spinmax=0.99, bins=100):
    fig = plt.figure()
    corner.hist2d(
        costs, spins, bins=bins, levels=[level], range=[[-1., 1.], [0., spinmax]], smooth=7, fig=fig
    )
    
    # Assume a single path
    cost_spin                = fig.axes[0].collections[0].get_paths()[0].vertices
    costcontour, spincontour = cost_spin[:, 0], cost_spin[:, 1]
    
    # Extract contour in range
    in_range_idxs = np.argwhere((costcontour > -1.) & (costcontour < 1.) & (spincontour > 0.))[:, 0]
    switch_points = np.argwhere(np.diff(in_range_idxs) > 1)
    start_idxs    = np.append(in_range_idxs[0], in_range_idxs[switch_points + 1])
    end_idxs      = np.append(in_range_idxs[switch_points], in_range_idxs[-1])
    
    # Collect contours
    costcontours, spincontours = [], []
    for s, e in zip(start_idxs, end_idxs):
        _costcontour, _spincontour = costcontour[s:e+1], spincontour[s:e+1]
        if s > 0:
            c1, c2       = costcontour[s-1], costcontour[s]
            a1, a2       = spincontour[s-1], spincontour[s]
            rs           = np.array([(0. - a1) / (a2 - a1), (-1. - c1) / (c2 - c1), (1. - c1) / (c2 - c1)])
            r            = np.max(rs[rs<=1.])
            _costcontour = np.append((c2 - c1) * r + c1, _costcontour)
            _spincontour = np.append((a2 - a1) * r + a1, _spincontour)
        if e + 1 < len(costcontour):
            c1, c2       = costcontour[e], costcontour[e+1]
            a1, a2       = spincontour[e], spincontour[e+1]
            rs           = np.array([(0. - a1) / (a2 - a1), (-1. - c1) / (c2 - c1), (1. - c1) / (c2 - c1)])
            r            = np.min(rs[rs>=0.])
            _costcontour = np.append(_costcontour, (c2 - c1) * r + c1)
            _spincontour = np.append(_spincontour, (a2 - a1) * r + a1) 
        # Append to list 
        costcontours.append(_costcontour)
        spincontours.append(_spincontour)
        
    return costcontours, spincontours

def spin_distribution_plots_with_contours(
    samples_highspin, 
    samples_lowspin, 
    label, 
    use_label         = True,
    level             = 0.9, 
    contour_linewidth = 1.0, 
    cmap              = None,
    color             = None, 
    colorbar          = True, 
    add_spin2_contour = True, 
    annotate          = True,
    fig               = None,
):
    """
    Generate spin distribution plots for a set of high and low spin samples
    
    Parameters
    ----------
    samples_highspin: dict
        dict of samples for high spin
    samples_lowspin: dict
        dict of samples for low spin 
    label: str
        the label corresponding to the set of samples
    use_label: bool
        add label to plot
    color: tuple
        tuple describing the color to be used for plotting
    level: float
        confidence interval of contours
    contour_linewidth: float
        linewidth of contours
    cmap: obj
        pass a color map
    color: str
        color for contour lines. also used for to generate cmap
        if cmap is None
    colorbar: bool
        
    add_spin2_contour: bool
        add second set of contours for low spin samples
    annotate:
        add annotations
    fig: object
        to pass own figure object
    """
    from matplotlib.ticker import FixedLocator, Formatter
    from matplotlib.projections import PolarAxes
    from matplotlib.transforms import Affine2D
    from matplotlib.patches import Wedge
    from matplotlib import patheffects as PathEffects
    from matplotlib.collections import PatchCollection
    from matplotlib.transforms import ScaledTranslation

    from mpl_toolkits.axisartist.grid_finder import MaxNLocator
    import mpl_toolkits.axisartist.floating_axes as floating_axes
    import mpl_toolkits.axisartist.angle_helper as angle_helper
    from matplotlib.colors import LinearSegmentedColormap

    if cmap is not None:
        from matplotlib import cm
        cmap = cm.get_cmap(cmap)
    else:
        cmap = utils.create_colormap_from_with_fixed_hue(label+"_color", color)

    # Get samples from high spin run
    spin1high  = samples_highspin["a_1"]
    spin2high  = samples_highspin["a_2"]
    theta1high = samples_highspin["cos_tilt_1"]
    theta2high = samples_highspin["cos_tilt_2"]
    
    # Get samples from low spin run
    spin1low   = samples_lowspin["a_1"]
    spin2low   = samples_lowspin["a_2"]
    theta1low  = samples_lowspin["cos_tilt_1"]
    theta2low  = samples_lowspin["cos_tilt_2"]

    pts              = np.array([spin1high, theta1high]).T
    selected_indices = np.random.choice(len(pts), len(pts) // 2, replace=False)
    kde_sel          = np.zeros(len(pts), dtype=bool)
    
    kde_sel[selected_indices] = True
    kde_pts          = pts[kde_sel]
    spin1            = bounded_2d_kde(
        kde_pts, xlow=0, xhigh=.99, ylow=-1, yhigh=1
    )
    pts              = np.array([spin2high, theta2high]).T
    selected_indices = np.random.choice(len(pts), len(pts) // 2, replace=False)
    kde_sel          = np.zeros(len(pts), dtype=bool)
    kde_sel[selected_indices] = True
    kde_pts          = pts[kde_sel]
    spin2            = bounded_2d_kde(
        kde_pts, xlow=0, xhigh=.99, ylow=-1, yhigh=1
    )

    n_bins    = 61 # 
    rs        = np.linspace(0, .99, n_bins)
    dr        = np.abs(rs[1] - rs[0])
    costs     = np.linspace(-1, 1, n_bins)
    dcost     = np.abs(costs[1] - costs[0])
    COSTS, RS = np.meshgrid(costs[:-1], rs[:-1])
    X         = np.arccos(COSTS) * 180 / np.pi + 90.
    Y         = RS

    scale     = np.exp(1.0)
    spin1_PDF = spin1(
        np.column_stack([RS.ravel() + dr / 2, COSTS.ravel() + dcost / 2]))
    spin2_PDF = spin2(
        np.column_stack([RS.ravel() + dr / 2, COSTS.ravel() + dcost / 2]))
    H1        = np.log(1.0 + scale * spin1_PDF)
    H2        = np.log(1.0 + scale * spin2_PDF)
    vmin      = 0.0
    vmax      = np.log(1.0 + scale * 3.024)

    rect      = 121

    tr        = Affine2D().translate(90, 0) + Affine2D().scale(np.pi / 180., 1.) + \
        PolarAxes.PolarTransform()

    grid_locator1   = angle_helper.LocatorD(7)
    tick_formatter1 = angle_helper.FormatterDMS()
    grid_locator2   = MaxNLocator(5)
    grid_helper     = floating_axes.GridHelperCurveLinear(
        tr, extremes=(0, 180, 0, .99),
        grid_locator1=grid_locator1,
        grid_locator2=grid_locator2,
        tick_formatter1=tick_formatter1,
        tick_formatter2=None
    )

    if fig is None:
        fig = plt.figure(figsize=(5,5))
    
    ax1 = floating_axes.FloatingSubplot(fig, rect, grid_helper=grid_helper)
    fig.add_subplot(ax1)

    ax1.axis["bottom"].toggle(all=False)
    ax1.axis["top"].toggle(all=True)
    ax1.axis["top"].major_ticks.set_tick_out(True)
    ax1.axis["top"].major_ticklabels.set()

    ax1.axis["top"].set_axis_direction("top")
    ax1.axis["top"].set_ticklabel_direction("+")

    ax1.axis["left"].major_ticks.set_tick_out(True)
    ax1.axis["left"].set_axis_direction("right")
    dx = 7.0 / 72.
    dy = 0 / 72.
    offset_transform = ScaledTranslation(dx, dy, fig.dpi_scale_trans)
    ax1.axis["left"].major_ticklabels.set(figure=fig,
                                          transform=offset_transform)
    ax1.axis["left"].major_ticklabels.set_pad(-4)
    
    patches = []
    colors  = []
    for x, y, h in zip(X.ravel(), Y.ravel(), H1.ravel()):
        cosx  = np.cos((x - 90) * np.pi / 180)
        cosxp = cosx + dcost
        xp    = np.arccos(cosxp)
        xp    = xp * 180. / np.pi + 90.
        patches.append(Wedge((0., 0.), y + dr, xp, x, width=dr))
        colors.append(h)

    p = PatchCollection(patches, cmap=cmap, edgecolors="face", zorder=10)
    p.set_clim(vmin, vmax)
    p.set_array(np.array(colors))
    ax1.add_collection(p)

    # Add contours
    ccontours, acontours = get_contour(theta1low, spin1low, level=level, spinmax=0.99)
    for ccontour, acontour in zip(ccontours, acontours):
        ax1.plot(-acontour * np.sqrt(1. - ccontour**2.), acontour * ccontour, zorder=20, color=color, linewidth=contour_linewidth, linestyle=(1.5, (4, 2.5)))
    ccontours, acontours = get_contour(theta1high, spin1high, level=level, spinmax=0.99)
    for ccontour, acontour in zip(ccontours, acontours):
        ax1.plot(-acontour * np.sqrt(1. - ccontour**2.), acontour * ccontour, zorder=20, color=color, linewidth=contour_linewidth)  
    
    # Spin 2
    rect = 122

    tr_rotate = Affine2D().translate(90, 0)
    tr_scale = Affine2D().scale(np.pi / 180., 1.)
    tr = tr_rotate + tr_scale + PolarAxes.PolarTransform()

    grid_locator1 = angle_helper.LocatorD(7)
    tick_formatter1 = angle_helper.FormatterDMS()

    grid_locator2 = MaxNLocator(5)

    grid_helper = floating_axes.GridHelperCurveLinear(
        tr, extremes=(0, 180, 0, .99),
        grid_locator1=grid_locator1,
        grid_locator2=grid_locator2,
        tick_formatter1=tick_formatter1,
        tick_formatter2=None)

    ax1 = floating_axes.FloatingSubplot(fig, rect, grid_helper=grid_helper)
    ax1.invert_xaxis()
    fig.add_subplot(ax1)

    # Label angles on the outside
    ax1.axis["bottom"].toggle(all=False)
    ax1.axis["top"].toggle(all=True)
    ax1.axis["top"].set_axis_direction("top")
    ax1.axis["top"].major_ticks.set_tick_out(True)
    ax1.axis["top"].major_ticklabels.set()

    # Remove radial labels
    ax1.axis["left"].major_ticks.set_tick_out(True)
    ax1.axis["left"].toggle(ticklabels=False)
    ax1.axis["left"].major_ticklabels.set_visible(False) ###

    # Also have radial ticks for the lower half of the right semidisk
    ax1.axis["right"].major_ticks.set_tick_out(True)
        
    patches = []
    colors = []
    for x, y, h in zip(X.ravel(), Y.ravel(), H2.ravel()):
        cosx  = np.cos((x - 90) * np.pi / 180)
        cosxp = cosx + dcost
        xp    = np.arccos(cosxp)
        xp    = xp * 180. / np.pi + 90.
        patches.append(Wedge((0., 0.), y + dr, xp, x, width=dr))
        colors.append(h)

    p = PatchCollection(patches, cmap=cmap, edgecolors="face", zorder=10)
    p.set_clim(vmin, vmax)
    p.set_array(np.array(colors))
    ax1.add_collection(p)

    # Add contour
    if add_spin2_contour:
        ccontours, acontours = get_contour(theta2low, spin2low, level=level, spinmax=0.05)
        for ccontour, acontour in zip(ccontours, acontours):
            ax1.plot(-acontour * np.sqrt(1. - ccontour**2.), acontour * ccontour, zorder=20, color=color, linewidth=contour_linewidth, linestyle=(1.5, (4, 2.5)))
        ccontours, acontours = get_contour(theta2high, spin2high, level=level, spinmax=0.99)
        for ccontour, acontour in zip(ccontours, acontours):
            ax1.plot(-acontour * np.sqrt(1. - ccontour**2.), acontour * ccontour, zorder=20, color=color, linewidth=contour_linewidth)
    
    # Event name top, spin labels bottom
    if label is not None:
        if use_label == True:
            title    = fig.text(0.17, 0.88, label, horizontalalignment="center")
            bbox_extra_artists = [title]
        else:
            bbox_extra_artists = None

    if colorbar:
        S1_label = plt.text(1.25, -1.2, r"$c\boldsymbol{S}_{1}/(Gm_1^2)$")
        S2_label = plt.text(-.04, -1.2, r"$c\boldsymbol{S}_{2}/(Gm_2^2)$")
        plt.subplots_adjust(wspace=0.295)

        # Add colorbar
        ax3  = plt.axes([0.22, 0.05, 0.55, 0.02])
        cbar = fig.colorbar(p, cax=ax3, orientation="horizontal",
                            pad=0.2, shrink=0.5)
    else:
        pass
        
    
    scale   = 1.
    
    if annotate:
        scale   = 1.0
        aux_ax2 = ax1.get_aux_axes(tr)
        scale   = 1
        txt     = aux_ax2.annotate("",
                       xy=(55, 1.158*scale), xycoords="data",
                       xytext=(35, 1.158*scale), textcoords="data",
                       arrowprops=dict(arrowstyle="->",
                                       color="k",
                                       shrinkA=2, shrinkB=2,
                                       patchA=None,
                                       patchB=None,
                                       connectionstyle="arc3,rad=-0.16")
                                  )
        txt.arrow_patch.set_path_effects([
        PathEffects.Stroke(linewidth=2, foreground="w"),
        PathEffects.Normal()])
        
        txt = aux_ax2.annotate("",
                       xy=(35, 0.55*scale), xycoords="data",
                       xytext=(150, 0.*scale), textcoords="data",
                       arrowprops=dict(arrowstyle="->",
                                       color="k",
                                       shrinkA=2, shrinkB=2,
                                       patchA=None,
                                       patchB=None), zorder=100)

        txt.arrow_patch.set_path_effects([
            PathEffects.Stroke(linewidth=0.3, foreground="k"),
            PathEffects.Normal()]
        )
    return fig, bbox_extra_artists
